module ZipWith where

import Level
open import Coinduction

open import Data.Nat
open import Data.Unit
open import Data.Product using (_×_; _,_)
open import Data.List using (List; []; _∷_; map)
open import Data.Stream using (Stream; _∷_; iterate; take; repeat)
  renaming (zipWith to zipWith₂; map to mapStream)

_⇒_ : List Set → Set → Set
[]      ⇒ C = C
(A ∷ L) ⇒ C = A → L ⇒ C

module ViaZip where

  Π : List Set → Set
  Π []      = ⊤
  Π (A ∷ L) = A × Π L

  curry : (L : List Set){C : Set} → (Π L → C) → L ⇒ C
  curry []      f = f tt
  curry (A ∷ L) f = λ a → curry L (λ as → f (a , as))

  uncurry : (L : List Set){C : Set} → (L ⇒ C) → Π L → C
  uncurry []      c tt       = c
  uncurry (A ∷ L) f (a , as) = uncurry L (f a) as

  -- zip₂ : {A B : Set} → Stream A → Stream B → Stream (A × B)
  -- zip₂ (a ∷ as) (b ∷ bs) = (a , b) ∷ ♯ zip₂ (♭ as) (♭ bs)

  zip : (L : List Set) → Π (map Stream L) → Stream (Π L)
  zip []      tt        = repeat tt
  zip (A ∷ L) (as , ss) = zipWith₂ _,_ as (zip L ss)

  zipWith : (L : List Set){C : Set}(f : L ⇒ C) → map Stream L ⇒ Stream C
  zipWith L f = curry (map Stream L) (λ ss → mapStream (uncurry L f) (zip L ss))

module Nonterm where

  -- induction on L
  zipWith' : (L : List Set){C : Set}(f : L ⇒ C)
    (k : map Stream L ⇒ Stream C) → map Stream L ⇒ Stream C
  zipWith' []      c cs = c ∷ ♯ cs
  zipWith' (A ∷ L) f k  = λ { (a ∷ s) → zipWith' L (f a) (k (♭ s)) }

  -- non-termination:
  zipWith : (L : List Set){C : Set}(f : L ⇒ C) → map Stream L ⇒ Stream C
  zipWith L f = zipWith' L f (zipWith L f)

-- induction on L
-- unconventional use of ∞
zipWith' : (L : List Set){C : Set}(f : L ⇒ C)
  (k : ∞ (map Stream L ⇒ Stream C)) → map Stream L ⇒ Stream C
zipWith' []      c cs = c ∷ cs
zipWith' (A ∷ L) f k  = λ { (a ∷ as) → zipWith' L (f a) (♯ (♭ k) (♭ as)) }

zipWith : (L : List Set){C : Set}(f : L ⇒ C) → map Stream L ⇒ Stream C
zipWith L f = zipWith' L f (♯ zipWith L f)

-- testing

nats : Stream ℕ
nats = iterate suc zero

squares : Stream ℕ
squares = zipWith (ℕ ∷ ℕ ∷ []) _*_ nats nats

test = take 6 squares
